<?php

/**
 * Locale for "Chinese (Simplified)".
 */
final class PhutilSimplifiedChineseLocale extends PhutilLocale {

  public function getLocaleCode() {
    return 'zh_Hans';
  }

  public function getLocaleName() {
    return pht('Chinese (Simplified)');
  }

}
