<?php

/**
 * The Spanish (Spain) Locale
 */
final class PhutilSpanishSpainLocale extends PhutilLocale {

  public function getLocaleCode() {
    return 'es_ES';
  }

  public function getLocaleName() {
    return pht('Spanish (Spain)');
  }

}
