<?php

/**
 * Locale for "French (France)".
 */
final class PhutilFrenchLocale extends PhutilLocale {

  public function getLocaleCode() {
    return 'fr_FR';
  }

  public function getLocaleName() {
    return pht('French (France)');
  }

}
