<?php

/**
 * The default English locale.
 */
final class PhutilUSEnglishLocale extends PhutilLocale {

  public function getLocaleCode() {
    return 'en_US';
  }

  public function getLocaleName() {
    return pht('English (US)');
  }

}
