<?php

/**
 * Locale for "German (Germany)".
 */
final class PhutilGermanLocale extends PhutilLocale {

  public function getLocaleCode() {
    return 'de_DE';
  }

  public function getLocaleName() {
    return pht('German (Germany)');
  }

}
