<?php

/**
 * The Korean (Republic of Korea) Locale
 */
final class PhutilKoreanLocale extends PhutilLocale {

  public function getLocaleCode() {
    return 'ko_KR';
  }

  public function getLocaleName() {
    return pht('Korean (Republic of Korea)');
  }

}
