<?php

/**
 * Locale for "Portuguese (Portugal)".
 */
final class PhutilPortuguesePortugalLocale extends PhutilLocale {

  public function getLocaleCode() {
    return 'pt_PT';
  }

  public function getLocaleName() {
    return pht('Portuguese (Portugal)');
  }

}
