<?php

/**
 * Locale for "Chinese (Traditional)".
 */
final class PhutilTraditionalChineseLocale extends PhutilLocale {

  public function getLocaleCode() {
    return 'zh_Hant';
  }

  public function getLocaleName() {
    return pht('Chinese (Traditional)');
  }

}
