<?php

/**
 * Locale for "Portuguese (Brazil)".
 */
final class PhutilPortugueseBrazilLocale extends PhutilLocale {

  public function getLocaleCode() {
    return 'pt_BR';
  }

  public function getLocaleName() {
    return pht('Portuguese (Brazil)');
  }

}
